package ru.dbotthepony.kstarbound.client.gl.shader

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.lwjgl.opengl.GL20.GL_FRAGMENT_SHADER
import org.lwjgl.opengl.GL20.GL_VERTEX_SHADER
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexAttributeType
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexAttributes
import ru.dbotthepony.kvector.api.IStruct4f
import ru.dbotthepony.kvector.arrays.Matrix3f
import ru.dbotthepony.kvector.vector.RGBAColor

class UberShader private constructor(fragment: GLShader, vertex: GLShader, attributes: VertexAttributes) : GLShaderProgram(listOf(fragment, vertex), attributes), GLShaderProgram.Regular {
	override var viewMatrix: Matrix3f by F3x3Uniform("viewMatrix")
	override var worldMatrix: Matrix3f by F3x3Uniform("worldMatrix")
	override var modelMatrix: Matrix3f by F3x3Uniform("modelMatrix")
	override var colorMultiplier: IStruct4f by F4Uniform("colorMultiplier")

	var texture0 by IUniform("texture0", VertexAttributeType.UV !in attributes)

	val builder = StreamVertexBuilder(attributes)

	init {
		viewMatrix = Matrix3f.identity()
		worldMatrix = Matrix3f.identity()
		modelMatrix = Matrix3f.identity()
		colorMultiplier = RGBAColor.WHITE
	}

	class Builder {
		private val directives = Object2ObjectArrayMap<String, String>()
		private val attributes = ObjectArraySet<VertexAttributeType>()

		init {
			attributes.add(VertexAttributeType.POSITION)
		}

		private var nextAttributeIndex = 1

		private fun attribute(type: VertexAttributeType, name: String = type.name): Builder {
			if (attributes.add(type)) {
				directives[name] = nextAttributeIndex.toString()
				nextAttributeIndex += type.type.width
			}

			return this
		}

		fun withTexture(): Builder {
			return attribute(VertexAttributeType.UV, "TEXTURE")
		}

		fun withColor(): Builder {
			return attribute(VertexAttributeType.COLOR)
		}

		fun withHueShift(): Builder {
			return attribute(VertexAttributeType.HUE_SHIFT)
		}

		fun build(): UberShader {
			val client = StarboundClient.current()

			val fragment = GLShader("#version 450\n" +
				                        directives.entries.joinToString("\n") { "#define ${it.key}" + (if (it.value != "") " " + it.value else "") } +
				                        fragment, GL_FRAGMENT_SHADER)

			val vertex = GLShader("#version 450\n" +
				                        directives.entries.joinToString("\n") { "#define ${it.key}" + (if (it.value != "") " " + it.value else "") } +
				                        vertex, GL_VERTEX_SHADER)

			val attributes = VertexAttributes.of(attributes)

			return UberShader(fragment, vertex, attributes)
		}
	}

	companion object {
		private val fragment by lazy { StarboundClient.readInternal("shaders/fragment/configurable.fsh") }
		private val vertex by lazy { StarboundClient.readInternal("shaders/vertex/configurable.vsh") }
	}
}
